// -*- mode: c++; c-file-style: "linux"; c-basic-offset: 2; indent-tabs-mode:
// nil -*-
//
//  Copyright (C) 2019      Gunter Königsmann <wxMaxima@physikbuch.de>
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
//
//  SPDX-License-Identifier: GPL-2.0+

/*! \file
  This file contains the logic needed in order to build a (hopefully)
  flicker-free main menu bar.
*/

#include "MainMenuBar.h"

MainMenuBar::MainMenuBar() : wxMenuBar() {}

void MainMenuBar::EnableItem(int id, bool enable) {
  if (IsEnabled(id) != enable)
    wxMenuBar::Enable(id, enable);
}

void MainMenuBar::CheckItem(int id, bool check) {
  if (IsChecked(id) != check)
    wxMenuBar::Check(id, check);
}
